package rendering

import (
	"context"
	"errors"
	"fmt"

	"orus.io/orus-io/rednerd/models"
)

// ErrDuplicateName is returned by EngineRegistry.Register if and engine with
// the same name is already registered
var ErrDuplicateName = errors.New("duplicate engine name")

// EngineTransitionInfo describes a
type EngineTransitionInfo struct {
	FromType string
	ToType   string
	Priority float64
}

// EngineInfo describes an engine capabilities
type EngineInfo struct {
	Type        string
	Description string
	Transitions []EngineTransitionInfo
}

// Engine is able to render a document
type Engine interface {
	Info() EngineInfo
	Render(ctx context.Context, document *models.Document, toType string) (*models.Document, error)
}

// RegisterEngine a new engine factory
func RegisterEngine(name string, engine Engine) error {
	return DefaultEngineRegistry.Register(name, engine)
}

// MustRegisterEngine registers an Engine and panics if it fails
func MustRegisterEngine(name string, engine Engine) {
	DefaultEngineRegistry.MustRegister(name, engine)
}

// EngineRegistry is a set of Engine, ready for use
type EngineRegistry struct {
	Engines map[string]Engine
}

// Register a Engine
func (s *EngineRegistry) Register(name string, engine Engine) error {
	if _, ok := s.Engines[name]; ok {
		return ErrDuplicateName
	}
	s.Engines[name] = engine
	return nil
}

// MustRegister a Engine
func (s *EngineRegistry) MustRegister(name string, engine Engine) {
	if err := s.Register(name, engine); err != nil {
		panic(err)
	}
}

// PipelineStep is a step in a pipeline
type PipelineStep struct {
	TargetType string
	Engine     Engine
}

// GetPipeline returns a list of engine for converting a document of type 'fromType'
// to type 'toType'
// TODO do a real implementation, not this hard-coded hack
func (s *EngineRegistry) GetPipeline(fromType, toType string) ([]PipelineStep, error) {
	if fromType == toType {
		return nil, nil
	}
	if fromType == "text/mjml" && toType == "application/pdf" {
		engine1, err := s.FindEngine(fromType, "text/html")
		if err != nil {
			return nil, err
		}
		engine2, err := s.FindEngine("text/html", toType)
		if err != nil {
			return nil, err
		}
		return []PipelineStep{
			{"text/html", engine1},
			{"application/pdf", engine2},
		}, nil
	}
	engine, err := s.FindEngine(fromType, toType)
	if err != nil {
		return nil, err
	}
	return []PipelineStep{
		{toType, engine},
	}, nil
}

// FindEngine finds the engine to do a transformation that has the highest priority
func (s *EngineRegistry) FindEngine(fromType, toType string) (Engine, error) {
	engines, err := s.FindEngines(fromType, toType)
	if err != nil {
		return nil, err
	}
	if len(engines) == 0 {
		return nil, fmt.Errorf("no engine found to convert from %s to %s", fromType, toType)
	}
	return engines[0], nil
}

// FindEngines finds engines that can do a given transformation. The result
// is sorted by priority, the highest first
func (s *EngineRegistry) FindEngines(fromType, toType string) ([]Engine, error) {
	type entry struct {
		engine   Engine
		priority float64
	}

	entries := []entry{}
	for _, engine := range s.Engines {
		for _, tr := range engine.Info().Transitions {
			if tr.FromType == fromType && tr.ToType == toType {
				entries = append(entries, entry{engine, tr.Priority})
			}
		}
	}
	var engines = make([]Engine, 0, len(entries))
	for _, entry := range entries {
		engines = append(engines, entry.engine)
	}
	return engines, nil
}

// NewEngineRegistry creates a EngineRegistry
func NewEngineRegistry() *EngineRegistry {
	return &EngineRegistry{
		Engines: make(map[string]Engine),
	}
}

var (
	// DefaultEngineRegistry is the default engine registry
	DefaultEngineRegistry = NewEngineRegistry()
)
