package rendering

import (
	"context"
	"errors"

	"orus.io/orus-io/rednerd/models"
)

// ErrUnknownTemplateLanguage is returned if no engine supports the given language
var ErrUnknownTemplateLanguage = errors.New("unknown template language")

// TemplateEngineInfo describes an engine capabilities
type TemplateEngineInfo struct {
	Type        string
	Language    string
	Description string
}

// Template is a parsed and loaded template, ready to be rendered
type Template interface {
	Render(ctx context.Context, record models.Record) (*models.Document, error)
}

// TemplateEngine is able to render a document
type TemplateEngine interface {
	Info() TemplateEngineInfo
	Load(ctx context.Context, template *models.Template) (Template, error)
}

// RegisterTemplateEngine a new engine factory
func RegisterTemplateEngine(name string, engine TemplateEngine) error {
	return DefaultTemplateEngineRegistry.Register(name, engine)
}

// MustRegisterTemplateEngine registers an TemplateEngine and panics if it fails
func MustRegisterTemplateEngine(name string, engine TemplateEngine) {
	DefaultTemplateEngineRegistry.MustRegister(name, engine)
}

// TemplateEngineRegistry is a set of TemplateEngine, ready for use
type TemplateEngineRegistry struct {
	TemplateEngines map[string]TemplateEngine
}

// Register a TemplateEngine
func (s *TemplateEngineRegistry) Register(name string, engine TemplateEngine) error {
	if _, ok := s.TemplateEngines[name]; ok {
		return ErrDuplicateName
	}
	s.TemplateEngines[name] = engine
	return nil
}

// MustRegister a TemplateEngine
func (s *TemplateEngineRegistry) MustRegister(name string, engine TemplateEngine) {
	if err := s.Register(name, engine); err != nil {
		panic(err)
	}
}

// Find finds a template engine by the language it supports
func (s *TemplateEngineRegistry) Find(language string) TemplateEngine {
	for _, engine := range s.TemplateEngines {
		if engine.Info().Language == language {
			return engine
		}
	}
	return nil
}

// Load a template with the proper engine
func (s *TemplateEngineRegistry) Load(ctx context.Context, template *models.Template) (Template, error) {
	engine := s.Find(template.Language)
	if engine == nil {
		return nil, ErrUnknownTemplateLanguage
	}
	return engine.Load(ctx, template)
}

// NewTemplateEngineRegistry creates a TemplateEngineRegistry
func NewTemplateEngineRegistry() *TemplateEngineRegistry {
	return &TemplateEngineRegistry{
		TemplateEngines: make(map[string]TemplateEngine),
	}
}

var (
	// DefaultTemplateEngineRegistry is the default engine registry
	DefaultTemplateEngineRegistry = NewTemplateEngineRegistry()
)
